configs_mimic_coral = {'dataset': 'mimic', 'regression': False, 'prediction_type': 'mortality', 'method': 'coral', 'device': 0, 'random_seed': 1, 'train_update_iter': 3000, 'lr': 5e-4, 'momentum': 0.9, 'weight_decay': 0.0, 'mini_batch_size': 128, 'reduced_train_prop': None, 'eval_fix': True, 'difficulty': False, 'split_time': 2011, 'eval_next_timestamps': 1, 'load_model': False, 'eval_all_timestamps': False, 'K': 1, 'lisa': False, 'lisa_intra_domain': False, 'mixup': False, 'lisa_start_time': 0, 'mix_alpha': 2.0, 'cut_mix': False, 'num_groups': 4, 'group_size': 1, 'non_overlapping': False, 'ewc_lambda': 1.0, 'gamma': 1.0, 'online': False, 'fisher_n': None, 'emp_FI': False, 'buffer_size': 1000, 'coral_lambda': 0.9, 'irm_lambda': 1.0, 'irm_penalty_anneal_iters': 0, 'si_c': 0.1, 'epsilon': 0.001, 'ssl_finetune_iter': 0, 'data_dir': './Data', 'log_dir': './checkpoints/', 'results_dir': './results', 'num_workers': 0}
configs_mimic_groupdro = {'dataset': 'mimic', 'regression': False, 'prediction_type': 'mortality', 'method': 'groupdro', 'device': 0, 'random_seed': 1, 'train_update_iter': 3000, 'lr': 5e-4, 'momentum': 0.9, 'weight_decay': 0.0, 'mini_batch_size': 128, 'reduced_train_prop': None, 'eval_fix': True, 'difficulty': False, 'split_time': 2011, 'eval_next_timestamps': 1, 'load_model': False, 'eval_all_timestamps': False, 'K': 1, 'lisa': False, 'lisa_intra_domain': False, 'mixup': False, 'lisa_start_time': 0, 'mix_alpha': 2.0, 'cut_mix': False, 'num_groups': 4, 'group_size': 1, 'non_overlapping': False, 'ewc_lambda': 1.0, 'gamma': 1.0, 'online': False, 'fisher_n': None, 'emp_FI': False, 'buffer_size': 1000, 'coral_lambda': 1.0, 'irm_lambda': 1.0, 'irm_penalty_anneal_iters': 0, 'si_c': 0.1, 'epsilon': 0.001, 'ssl_finetune_iter': 0, 'data_dir': './Data', 'log_dir': './checkpoints/', 'results_dir': './results', 'num_workers': 0}
configs_mimic_irm = {'dataset': 'mimic', 'regression': False, 'prediction_type': 'mortality', 'method': 'irm', 'device': 0, 'random_seed': 1, 'train_update_iter': 3000, 'lr': 5e-4, 'momentum': 0.9, 'weight_decay': 0.0, 'mini_batch_size': 128, 'reduced_train_prop': None, 'eval_fix': True, 'difficulty': False, 'split_time': 2011, 'eval_next_timestamps': 1, 'load_model': False, 'eval_all_timestamps': False, 'K': 1, 'lisa': False, 'lisa_intra_domain': False, 'mixup': False, 'lisa_start_time': 0, 'mix_alpha': 2.0, 'cut_mix': False, 'num_groups': 4, 'group_size': 1, 'non_overlapping': False, 'ewc_lambda': 1.0, 'gamma': 1.0, 'online': False, 'fisher_n': None, 'emp_FI': False, 'buffer_size': 1000, 'coral_lambda': 1.0, 'irm_lambda': 1.0, 'irm_penalty_anneal_iters': 0, 'si_c': 0.1, 'epsilon': 0.001, 'ssl_finetune_iter': 0, 'data_dir': './Data', 'log_dir': './checkpoints/', 'results_dir': './results', 'num_workers': 0}
configs_mimic_erm = {'dataset': 'mimic', 'regression': False, 'prediction_type': 'mortality', 'method': 'erm', 'device': 0, 'random_seed': 1, 'train_update_iter': 3000, 'lr': 5e-4, 'momentum': 0.9, 'weight_decay': 0.0, 'mini_batch_size': 128, 'reduced_train_prop': None, 'eval_fix': True, 'difficulty': False, 'split_time': 2011, 'eval_next_timestamps': 1, 'load_model': False, 'eval_all_timestamps': False, 'K': 1, 'lisa': False, 'lisa_intra_domain': False, 'mixup': False, 'lisa_start_time': 0, 'mix_alpha': 2.0, 'cut_mix': False, 'num_groups': 4, 'group_size': 1, 'non_overlapping': False, 'ewc_lambda': 1.0, 'gamma': 1.0, 'online': False, 'fisher_n': None, 'emp_FI': False, 'buffer_size': 1000, 'coral_lambda': 1.0, 'irm_lambda': 1.0, 'irm_penalty_anneal_iters': 0, 'si_c': 0.1, 'epsilon': 0.001, 'ssl_finetune_iter': 0, 'data_dir': './Data', 'log_dir': './checkpoints/', 'results_dir': './results', 'num_workers': 0}
configs_mimic_erm_diff = {'dataset': 'mimic', 'regression': False, 'prediction_type': 'mortality', 'method': 'erm', 'device': 0, 'random_seed': 1, 'train_update_iter': 3000, 'lr': 5e-4, 'momentum': 0.9, 'weight_decay': 0.0, 'mini_batch_size': 128, 'reduced_train_prop': None, 'eval_fix': True, 'difficulty': True, 'split_time': 2011, 'eval_next_timestamps': 1, 'load_model': False, 'eval_all_timestamps': False, 'K': 1, 'lisa': False, 'lisa_intra_domain': False, 'mixup': False, 'lisa_start_time': 0, 'mix_alpha': 2.0, 'cut_mix': False, 'num_groups': 4, 'group_size': 1, 'non_overlapping': False, 'ewc_lambda': 1.0, 'gamma': 1.0, 'online': False, 'fisher_n': None, 'emp_FI': False, 'buffer_size': 1000, 'coral_lambda': 1.0, 'irm_lambda': 1.0, 'irm_penalty_anneal_iters': 0, 'si_c': 0.1, 'epsilon': 0.001, 'ssl_finetune_iter': 0, 'data_dir': './Data', 'log_dir': './checkpoints/', 'results_dir': './results', 'num_workers': 0}
configs_mimic_erm_lisa = {'dataset': 'mimic', 'regression': False, 'prediction_type': 'mortality', 'method': 'erm', 'device': 0, 'random_seed': 1, 'train_update_iter': 3000, 'lr': 5e-4, 'momentum': 0.9, 'weight_decay': 0.0, 'mini_batch_size': 128, 'reduced_train_prop': None, 'eval_fix': True, 'difficulty': False, 'split_time': 2011, 'eval_next_timestamps': 1, 'load_model': False, 'eval_all_timestamps': False, 'K': 1, 'lisa': True, 'lisa_intra_domain': False, 'mixup': False, 'lisa_start_time': 0, 'mix_alpha': 2.0, 'cut_mix': False, 'num_groups': 4, 'group_size': 1, 'non_overlapping': False, 'ewc_lambda': 1.0, 'gamma': 1.0, 'online': False, 'fisher_n': None, 'emp_FI': False, 'buffer_size': 1000, 'coral_lambda': 1.0, 'irm_lambda': 1.0, 'irm_penalty_anneal_iters': 0, 'si_c': 0.1, 'epsilon': 0.001, 'ssl_finetune_iter': 0, 'data_dir': './Data', 'log_dir': './checkpoints/', 'results_dir': './results', 'num_workers': 0}
configs_mimic_erm_mixup  = {'dataset': 'mimic', 'regression': False, 'prediction_type': 'mortality', 'method': 'erm', 'device': 0, 'random_seed': 1, 'train_update_iter': 3000, 'lr': 5e-4, 'momentum': 0.9, 'weight_decay': 0.0, 'mini_batch_size': 128, 'reduced_train_prop': None, 'eval_fix': True, 'difficulty': False, 'split_time': 2011, 'eval_next_timestamps': 1, 'load_model': False, 'eval_all_timestamps': False, 'K': 1, 'lisa': False, 'lisa_intra_domain': False, 'mixup': True, 'lisa_start_time': 0, 'mix_alpha': 2.0, 'cut_mix': False, 'num_groups': 4, 'group_size': 1, 'non_overlapping': False, 'ewc_lambda': 1.0, 'gamma': 1.0, 'online': False, 'fisher_n': None, 'emp_FI': False, 'buffer_size': 1000, 'coral_lambda': 1.0, 'irm_lambda': 1.0, 'irm_penalty_anneal_iters': 0, 'si_c': 0.1, 'epsilon': 0.001, 'ssl_finetune_iter': 0, 'data_dir': './Data', 'log_dir': './checkpoints/', 'results_dir': './results', 'num_workers': 0}
configs_mimic_agem = {'dataset': 'mimic', 'regression': False, 'prediction_type': 'mortality', 'method': 'agem', 'device': 0, 'random_seed': 1, 'train_update_iter': 3000, 'lr': 5e-4, 'momentum': 0.9, 'weight_decay': 0.0, 'mini_batch_size': 128, 'reduced_train_prop': None, 'eval_fix': True, 'difficulty': False, 'split_time': 2011, 'eval_next_timestamps': 1, 'load_model': False, 'eval_all_timestamps': False, 'K': 1, 'lisa': False, 'lisa_intra_domain': False, 'mixup': False, 'lisa_start_time': 0, 'mix_alpha': 2.0, 'cut_mix': False, 'num_groups': 4, 'group_size': 1, 'non_overlapping': False, 'ewc_lambda': 1.0, 'gamma': 1.0, 'online': False, 'fisher_n': None, 'emp_FI': False, 'buffer_size': 1000, 'coral_lambda': 1.0, 'irm_lambda': 1.0, 'irm_penalty_anneal_iters': 0, 'si_c': 0.1, 'epsilon': 0.001, 'ssl_finetune_iter': 0, 'data_dir': './Data', 'log_dir': './checkpoints/', 'results_dir': './results', 'num_workers': 0}
configs_mimic_ewc = {'dataset': 'mimic', 'regression': False, 'prediction_type': 'mortality', 'method': 'ewc', 'device': 0, 'random_seed': 1, 'train_update_iter': 3000, 'lr': 5e-4, 'momentum': 0.9, 'weight_decay': 0.0, 'mini_batch_size': 128, 'reduced_train_prop': None, 'eval_fix': True, 'difficulty': False, 'split_time': 2011, 'eval_next_timestamps': 1, 'load_model': False, 'eval_all_timestamps': False, 'K': 1, 'lisa': False, 'lisa_intra_domain': False, 'mixup': False, 'lisa_start_time': 0, 'mix_alpha': 2.0, 'cut_mix': False, 'num_groups': 4, 'group_size': 1, 'non_overlapping': False, 'ewc_lambda': 0.5, 'gamma': 1.0, 'online': True, 'fisher_n': None, 'emp_FI': False, 'buffer_size': 1000, 'coral_lambda': 1.0, 'irm_lambda': 1.0, 'irm_penalty_anneal_iters': 0, 'si_c': 0.1, 'epsilon': 0.001, 'ssl_finetune_iter': 0, 'data_dir': './Data', 'log_dir': './checkpoints/', 'results_dir': './results', 'num_workers': 0}
configs_mimic_ft = {'dataset': 'mimic', 'regression': False, 'prediction_type': 'mortality', 'method': 'ft', 'device': 0, 'random_seed': 1, 'train_update_iter': 3000, 'lr': 5e-4, 'momentum': 0.9, 'weight_decay': 0.0, 'mini_batch_size': 128, 'reduced_train_prop': None, 'eval_fix': True, 'difficulty': False, 'split_time': 2011, 'eval_next_timestamps': 1, 'load_model': False, 'eval_all_timestamps': False, 'K': 1, 'lisa': False, 'lisa_intra_domain': False, 'mixup': False, 'lisa_start_time': 0, 'mix_alpha': 2.0, 'cut_mix': False, 'num_groups': 4, 'group_size': 1, 'non_overlapping': False, 'ewc_lambda': 1.0, 'gamma': 1.0, 'online': False, 'fisher_n': None, 'emp_FI': False, 'buffer_size': 1000, 'coral_lambda': 1.0, 'irm_lambda': 1.0, 'irm_penalty_anneal_iters': 0, 'si_c': 0.1, 'epsilon': 0.001, 'ssl_finetune_iter': 0, 'data_dir': './Data', 'log_dir': './checkpoints/', 'results_dir': './results', 'num_workers': 0}
configs_mimic_si = {'dataset': 'mimic', 'regression': False, 'prediction_type': 'mortality', 'method': 'si', 'device': 0, 'random_seed': 1, 'train_update_iter': 3000, 'lr': 5e-4, 'momentum': 0.9, 'weight_decay': 0.0, 'mini_batch_size': 128, 'reduced_train_prop': None, 'eval_fix': True, 'difficulty': False, 'split_time': 2011, 'eval_next_timestamps': 1, 'load_model': False, 'eval_all_timestamps': False, 'K': 1, 'lisa': False, 'lisa_intra_domain': False, 'mixup': False, 'lisa_start_time': 0, 'mix_alpha': 2.0, 'cut_mix': False, 'num_groups': 4, 'group_size': 1, 'non_overlapping': False, 'ewc_lambda': 1.0, 'gamma': 1.0, 'online': False, 'fisher_n': None, 'emp_FI': False, 'buffer_size': 1000, 'coral_lambda': 1.0, 'irm_lambda': 1.0, 'irm_penalty_anneal_iters': 0, 'si_c': 0.1, 'epsilon': 0.001, 'ssl_finetune_iter': 0, 'data_dir': './Data', 'log_dir': './checkpoints/', 'results_dir': './results', 'num_workers': 0}
configs_mimic_swa = {'dataset': 'mimic', 'regression': False, 'prediction_type': 'mortality', 'method': 'swa', 'device': 0, 'random_seed': 1, 'train_update_iter': 3000, 'lr': 5e-4, 'momentum': 0.9, 'weight_decay': 0.0, 'mini_batch_size': 128, 'reduced_train_prop': None, 'eval_fix': True, 'difficulty': False, 'split_time': 2011, 'eval_next_timestamps': 1, 'load_model': False, 'eval_all_timestamps': False, 'K': 1, 'lisa': False, 'lisa_intra_domain': False, 'mixup': False, 'lisa_start_time': 0, 'mix_alpha': 2.0, 'cut_mix': False, 'num_groups': 4, 'group_size': 1, 'non_overlapping': False, 'ewc_lambda': 1.0, 'gamma': 1.0, 'online': False, 'fisher_n': None, 'emp_FI': False, 'buffer_size': 1000, 'coral_lambda': 1.0, 'irm_lambda': 1.0, 'irm_penalty_anneal_iters': 0, 'si_c': 0.1, 'epsilon': 0.001, 'ssl_finetune_iter': 0, 'data_dir': './Data', 'log_dir': './checkpoints/', 'results_dir': './results', 'num_workers': 0}